from __future__ import annotations

import re
from typing import ClassVar, Optional, TYPE_CHECKING

import attr

from tsg.models import AuthorRef, SeriesRef, Book

if TYPE_CHECKING:
    from yarl import URL
    import bs4.element


@attr.s
class BookPageParser:
    _HREF_RE_TITLE: ClassVar[re.Pattern] = re.compile(r'/books/(?P<book_id>[\w\d\-]+$)')
    _HREF_RE_SERIES: ClassVar[re.Pattern] = re.compile(r'/series/(?P<series_id>[\w\d\-]+$)')
    _HREF_RE_AUTHOR: ClassVar[re.Pattern] = re.compile(r'/authors/(?P<author_id>[\w\d\-]+$)')

    _base_url: URL = attr.ib(kw_only=True)

    def parse_cover_url(self, book_cover_div: Optional[bs4.element.Tag]) -> Optional[str]:
        cover_url: Optional[str] = None
        if book_cover_div is not None:
            cover_url = book_cover_div.find('img').attrs['src']
        return cover_url

    def parse_book_pane(self, book_pane: bs4.element.Tag) -> Book:
        book_cover_div = book_pane.find('div', 'book-cover')
        cover_url = self.parse_cover_url(book_cover_div)

        title_author_series = book_pane.find('div', 'book-title-author-and-series')
        # Title and ID
        title_node = title_author_series.find('a', href=self._HREF_RE_TITLE)
        book_rel_url = title_node.get('href')
        book_id = self._HREF_RE_TITLE.match(book_rel_url).group('book_id')  # type: ignore
        book_title = title_node.get_text()
        book_full_url = str(self._base_url / book_rel_url.strip('/'))

        # Series
        series_node = title_author_series.find('a', href=self._HREF_RE_SERIES)
        series: Optional[SeriesRef] = None
        if series_node is not None:
            series_id = self._HREF_RE_SERIES.match(series_node.get('href')).group('series_id')  # type: ignore
            series_name = series_node.get_text()
            series = SeriesRef(id=series_id, name=series_name)

        # Authors
        authors: list[AuthorRef] = []
        author_node_list = title_author_series.find_all('a', href=self._HREF_RE_AUTHOR)
        for author_node in author_node_list:
            author_id = self._HREF_RE_AUTHOR.match(author_node.get('href')).group('author_id')  # type: ignore
            author_name = author_node.get_text()
            authors.append(AuthorRef(id=author_id, name=author_name))

        book = Book(
            id=book_id,
            title=book_title,
            series=series,
            authors=authors,
            url=book_full_url,
            cover_url=cover_url,
        )
        return book

    def parse_descr_book_pane(self, descr_book_pane: bs4.element.Tag) -> str:
        h4 = descr_book_pane.find('h4')
        return h4.get_text().strip()



from __future__ import annotations

from typing import Any

import attr
from bs4 import BeautifulSoup
from yarl import URL

from tsg.handlers.base import BaseHandler
from tsg.models import Book
from tsg.handlers.book_common import BookPageParser


@attr.s
class BrowseHandler(BaseHandler[list[Book]]):
    _book_page_parser: BookPageParser = attr.ib(init=False)

    def __attrs_post_init__(self) -> None:
        self._book_page_parser = BookPageParser(base_url=self._base_url)

    def make_url(self, text: str = '', **kwargs: Any) -> URL:
        return (self._base_url / 'browse').with_query(search_term=text)

    def parse_body(self, body: str) -> list[Book]:
        soup = BeautifulSoup(body, 'html.parser')
        raw_items = soup.find_all('div', 'book-pane')
        result: list[Book] = []
        for book_pane in raw_items:
            book = self._book_page_parser.parse_book_pane(book_pane)
            result.append(book)

        return result



from typing import Union

import attr
import requests
from aiohttp import ClientSession
from yarl import URL

from tsg.models import Book
from tsg.handlers.browse import BrowseHandler
from tsg.handlers.book import BookHandler


BASE_URL = 'https://app.thestorygraph.com'


@attr.s
class TSGClientBase:
    _base_url: URL = attr.ib(default=URL(BASE_URL))
    # Internal
    _browse_handler: BrowseHandler = attr.ib(init=False)
    _book_handler: BookHandler = attr.ib(init=False)

    def __attrs_post_init__(self):
        self._browse_handler = BrowseHandler(base_url=self._base_url)
        self._book_handler = BookHandler(base_url=self._base_url)


@attr.s
class SyncTSGClient(TSGClientBase):
    def _get(self, url: Union[str, URL]) -> str:
        return requests.get(str(url)).text

    def get_browse(self, text: str) -> list[Book]:
        return self._browse_handler.parse_body(
            self._get(self._browse_handler.make_url(text=text))
        )

    def get_book(self, id: str) -> Book:
        return self._book_handler.parse_body(
            self._get(self._book_handler.make_url(id=id))
        )


@attr.s
class AsyncTSGClient(TSGClientBase):
    async def _get(self, url: Union[str, URL]) -> str:
        async with ClientSession() as session:
            async with session.get(url) as response:
                return await response.text()

    async def get_browse(self, text: str) -> list[Book]:
        return self._browse_handler.parse_body(
            await self._get(self._browse_handler.make_url(text=text))
        )

    async def get_book(self, id: str) -> Book:
        return self._book_handler.parse_body(
            await self._get(self._book_handler.make_url(id=id))
        )



from __future__ import annotations

from typing import Any, Optional

import attr


@attr.s(frozen=True)
class AuthorRef:
    id: str = attr.ib(kw_only=True)
    name: str = attr.ib(kw_only=True)


@attr.s(frozen=True)
class SeriesRef:
    id: str = attr.ib(kw_only=True)
    name: str = attr.ib(kw_only=True)


@attr.s(frozen=True)
class Book:
    id: str = attr.ib(kw_only=True)
    title: str = attr.ib(kw_only=True)
    series: Optional[SeriesRef] = attr.ib(kw_only=True)
    authors: list[AuthorRef] = attr.ib(kw_only=True)
    url: str = attr.ib(kw_only=True)
    cover_url: Optional[str] = attr.ib(kw_only=True, default=None)
    description: str = attr.ib(kw_only=True, default='')

    @property
    def author_names(self) -> list[str]:
        return [author.name for author in self.authors]

    def clone(self, **kwargs: Any) -> Book:
        return attr.evolve(self, **kwargs)



from __future__ import annotations

import abc
from typing import Any, Generic, TypeVar, TYPE_CHECKING

import attr

if TYPE_CHECKING:
    from yarl import URL


_HANDLER_RES_TV = TypeVar('_HANDLER_RES_TV')


@attr.s
class BaseHandler(abc.ABC, Generic[_HANDLER_RES_TV]):
    _base_url: URL = attr.ib(kw_only=True)

    @abc.abstractmethod
    def make_url(self, **kwargs: Any) -> URL:
        raise NotImplementedError

    @abc.abstractmethod
    def parse_body(self, body: str) -> _HANDLER_RES_TV:
        raise NotImplementedError



from __future__ import annotations

from typing import Any, Optional

import attr
from bs4 import BeautifulSoup
from yarl import URL

from tsg import exc
from tsg.handlers.base import BaseHandler
from tsg.models import Book
from tsg.handlers.book_common import BookPageParser


@attr.s
class BookHandler(BaseHandler[Book]):
    _book_page_parser: BookPageParser = attr.ib(init=False)

    def __attrs_post_init__(self) -> None:
        self._book_page_parser = BookPageParser(base_url=self._base_url)

    def make_url(self, id: str = '', **kwargs: Any) -> URL:
        assert id
        return self._base_url / 'books' / id

    def parse_body(self, body: str) -> Book:
        soup = BeautifulSoup(body, 'html.parser')
        book_panes = soup.find_all('div', 'book-pane')
        book: Optional[Book] = None
        description = ''
        for book_pane in book_panes:
            title_author_series = book_pane.find('div', 'book-title-author-and-series')
            if title_author_series is not None:
                book = self._book_page_parser.parse_book_pane(book_pane)
            elif book is not None:
                # Main book pane has already been found
                text = book_pane.get_text().strip()
                if text.startswith('Description'):
                    description = text.split('\n', 1)[1].strip()

            if description:
                # Everything has been found
                break

        if book is None:
            raise exc.FailedToParse('Failed to parse book')

        book_cover_div = soup.find('div', 'book-cover')
        cover_url = self._book_page_parser.parse_cover_url(book_cover_div)
        book = book.clone(description=description, cover_url=cover_url)
        return book
